# import inspect
# from collections.abc import Callable
# from typing import TypedDict

# from pydantic import BaseModel
# from typing_extensions import Self


# class Dependency:
#     def __init__(self: Self, func: Callable) -> None:
#         self.func = func


# def is_dependency(arg) -> bool:
#     return isinstance(arg.param, Dependency)


# class DependencySchema(TypedDict):
#     arg_name: str
#     func: Callable
#     dependencies: dict[Callable, list["DependencySchema"]]


# class FunctionDependencies(BaseModel):
#     dependency_tree: dict[str, DependencySchema]


# def get_dependencies(func: Callable) -> FunctionDependencies:
#     dependencies = {}
#     for arg in inspect.signature(func).parameters.values():
#         if is_dependency(arg):
#             dependencies[arg.name] = DependencySchema(
#                 arg_name=arg.name,
#                 func=arg.default.func,
#                 dependencies=get_dependencies(arg.default.func).dependency_tree,
#             )

#     return FunctionDependencies(dependency_tree=dependencies)



import json
from enum import Enum
from typing import Any

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore

from aws_spy.core.encoders import JSONEncoder
from aws_spy.core.responses import BaseResponseSPY
from aws_spy.core.schemas import SpyRoute


class ContentType(str, Enum):
    JSON = "application/json"
    XML = "application/xml"
    PLAIN = "text/plain"
    HTML = "text/html"


class JSONResponse(BaseResponseSPY):
    def __init__(
        self: Self,
        data: dict[str, Any] | BaseModel,
        additional_headers: dict[str, str] | None = None,
    ) -> None:
        self.data = data
        self.additional_headers = additional_headers
        self.route: SpyRoute | None = None

    @property
    def response(self: Self) -> dict[str, Any]:
        headers = {
            "Content-Type": ContentType.JSON.value,
        }
        if self.additional_headers is not None:
            headers.update(self.additional_headers)

        if self.route is not None and self.route.response_class is not None and isinstance(self.data, dict):
            self.data = self.route.response_class.model_validate(self.data)

        if (
            self.route is not None
            and self.route.response_class is not None
            and isinstance(self.data, BaseModel)
            and not isinstance(self.data, self.route.response_class)
        ):
            self.data = self.route.response_class.model_validate(self.data.model_dump())

        if isinstance(self.data, BaseModel):
            self.data = self.data.model_dump()

        if self.route is None or self.route.status_code is None:
            status_code = 200
        else:
            status_code = self.route.status_code

        return {
            "statusCode": status_code,
            "body": json.dumps(self.data, cls=JSONEncoder, ensure_ascii=False),
            "headers": headers,
        }


class RAWResponse(BaseResponseSPY):
    def __init__(self: Self, response: dict[str, Any]) -> None:
        self.response_ = response

    @property
    def response(self: Self) -> dict[str, Any]:
        return self.response_


class ErrorResponse(BaseResponseSPY):
    def __init__(
        self: Self,
        errors: list[str],
        status_code: int = 400,
        additional_headers: dict[str, str] | None = None,
    ) -> None:
        self.errors = errors
        self.status_code = status_code
        self.additional_headers = additional_headers

    @property
    def response(self: Self) -> dict[str, Any]:
        headers = {
            "Content-Type": ContentType.JSON.value,
        }
        if self.additional_headers is not None:
            headers.update(self.additional_headers)

        return {
            "statusCode": self.status_code,
            "body": json.dumps(
                {"errors": [{"message": error} for error in self.errors]},
                cls=JSONEncoder,
                ensure_ascii=False,
            ),
            "headers": headers,
        }



from enum import Enum
from typing import Union
from uuid import UUID

import pytest

from aws_spy import Header, Path, Query, SpyAPI
from aws_spy.core.exceptions import RouteDefinitionError
from aws_spy.core.schemas import Methods, ParamSchema


class ExampleEnum(str, Enum):
    example = "example"


METHODS = list(Methods)


@pytest.mark.parametrize("method", METHODS)
def test_path_param_not_found(app: SpyAPI, method: Methods) -> None:
    app_method = getattr(app, method)
    path = "/path_without_user_id"

    with pytest.raises(
        RouteDefinitionError,
        match=(f'Your user_id path parameter is missing in {method.upper()} method on "{path}" path!'),
    ):

        @app_method(path, "lambda")
        def handler(user_id: str = Path()) -> None:
            ...


@pytest.mark.parametrize("method", METHODS)
def test_same_param_names(app: SpyAPI, method: Methods) -> None:
    app_method = getattr(app, method)
    path = "path"

    with pytest.raises(
        RouteDefinitionError,
        match='handler expects two same header params: "user_id"!',
    ):

        @app_method(path, "lambda")
        def handler(user_id: str = Header(), something: str = Header("user_id")) -> None:
            ...


@pytest.mark.parametrize("method", METHODS)
@pytest.mark.parametrize(
    ["path", "param", "route_param"],
    [
        ["/user_id", Query(), "query_params"],
        ["/{user_id}", Path(), "path_params"],
        ["/user_id", Header(), "header_params"],
    ],
)
@pytest.mark.parametrize(
    ["param_annotation", "expected_annotation", "is_required", "enum"],
    [
        [str, str, True, None],
        [int, int, True, None],
        [ExampleEnum, ExampleEnum, True, ["example"]],
        [UUID, UUID, True, None],
        [Union[str, None], str, False, None],
        [Union[ExampleEnum, None], ExampleEnum, False, ["example"]],
        [Union[str, int, ExampleEnum, None], str, False, None],  # only first arg
    ],
)
def test_params(
    app: SpyAPI,
    method: Methods,
    param,
    path: str,
    route_param: str,
    param_annotation: type,
    expected_annotation: type,
    is_required: bool,
    enum: list[str] | None,
) -> None:
    app_method = getattr(app, method.value)
    expected_param = ParamSchema(
        name="user_id",
        arg_name="user_id",
        in_=param,
        annotation=expected_annotation,
        is_required=is_required,
        enum=enum,
    )

    @app_method(path, "lambda")
    def handler(user_id: param_annotation = param) -> None:  # type: ignore
        ...

    assert len(app.routes) == 1
    route = app.routes[path][method]
    params = getattr(route, route_param)
    assert len(params) == 1
    assert params[0] == expected_param



import json
from typing import Any

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore

from aws_spy.core.schemas import LH, Methods


class Response(BaseModel):
    body: str
    status_code: int
    raw: dict[str, Any]

    @property
    def json_body(self: Self) -> dict[str, Any]:
        try:
            return json.loads(self.body)
        except json.JSONDecodeError:
            return {}


class TestClient:
    __test__ = False

    @classmethod
    def post(  # type: ignore
        cls: type[Self],
        handler: LH,
        *,
        body: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        path_params: dict[str, str] | None = None,
    ) -> Response:
        return cls._call(
            handler,
            cls._build_event(
                method=Methods.POST,
                body=body,
                headers=headers,
                path_params=path_params,
                query_params=query_params,
            ),
        )

    @classmethod
    def delete(  # type: ignore
        cls: type[Self],
        handler: LH,
        *,
        body: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        path_params: dict[str, str] | None = None,
    ) -> Response:
        return cls._call(
            handler,
            cls._build_event(
                method=Methods.DELETE,
                body=body,
                headers=headers,
                path_params=path_params,
                query_params=query_params,
            ),
        )

    @classmethod
    def patch(  # type: ignore
        cls: type[Self],
        handler: LH,
        *,
        body: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        path_params: dict[str, str] | None = None,
    ) -> Response:
        return cls._call(
            handler,
            cls._build_event(
                method=Methods.PATCH,
                body=body,
                headers=headers,
                path_params=path_params,
                query_params=query_params,
            ),
        )

    @classmethod
    def put(  # type: ignore
        cls: type[Self],
        handler: LH,
        *,
        body: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        path_params: dict[str, str] | None = None,
    ) -> Response:
        return cls._call(
            handler,
            cls._build_event(
                method=Methods.PUT,
                body=body,
                headers=headers,
                path_params=path_params,
                query_params=query_params,
            ),
        )

    @classmethod
    def get(  # type: ignore
        cls: type[Self],
        handler: LH,
        *,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        path_params: dict[str, str] | None = None,
    ) -> Response:
        return cls._call(
            handler,
            cls._build_event(
                method=Methods.GET,
                headers=headers,
                query_params=query_params,
                path_params=path_params,
            ),
        )

    @staticmethod
    def _call(handler: LH, event: dict[str, Any]) -> Response:
        response = handler(event, None)
        return Response(status_code=response["statusCode"], raw=response, body=response["body"])

    @staticmethod
    def _build_event(
        *,
        method: Methods,
        body: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, str] | None,
        path_params: dict[str, str] | None,
    ) -> dict[str, Any]:
        if body is None:
            body = {}
        base_headers = {
            "content-type": "application/json",
        }
        if headers is not None:
            base_headers.update(headers)

        return {
            "httpMethod": method.value,
            "headers": base_headers,
            "queryStringParameters": query_params,
            "pathParameters": path_params,
            "body": json.dumps(body),
        }



import json
from datetime import date, datetime
from decimal import Decimal
from uuid import UUID

from aws_spy.core.encoders import JSONEncoder


def test_json_encoder() -> None:
    uid = UUID("a2618752-dc19-4f40-adf5-6173d419f2ed")
    obj = {
        "decimal": Decimal(0.55),
        "datetime": datetime(2023, 5, 10),
        "date": date(2022, 5, 10),
        "uuid": uid,
        "default": 1,
    }
    encoded = json.dumps(obj, cls=JSONEncoder)
    assert encoded == (
        '{"decimal": 0.55, "datetime": "2023-05-10T00:00:00", "date": "2022-05-10", '
        '"uuid": "a2618752-dc19-4f40-adf5-6173d419f2ed", "default": 1}'
    )



import importlib
import sys
from pathlib import Path

from aws_spy import SpyAPI


class LoadAppFromStringError(Exception):
    ...


def load_app_from_string(import_string: str) -> SpyAPI:
    sys.path.append(str(Path().resolve()))

    module_string, _, app_name = import_string.partition(":")
    if not module_string or not app_name:
        msg = 'Path must be in format "<module>:<app_name>"'
        raise LoadAppFromStringError(msg)

    try:
        module = importlib.import_module(module_string)
    except ImportError as exc:
        if exc.name != module_string:
            raise exc from None
        msg = f'Could not import "{module_string}"'
        raise LoadAppFromStringError(msg) from None

    try:
        return getattr(module, app_name)
    except AttributeError:
        msg = f'Attribute "{app_name}" not found in "{module_string}" module'
        raise LoadAppFromStringError(msg) from None



import inspect
import os
from collections.abc import Callable
from enum import Enum
from pathlib import Path
from typing import Any, Literal, TypeVar

from pydantic import BaseModel, Field, field_validator, model_validator
from typing_extensions import Self  # type: ignore

from aws_spy.core.exceptions import RouteDefinitionError
from aws_spy.core.schemas_utils import (
    ParamSchema,
    get_path_param_names,
    resolve_handler_args,
)

# from aws_spy.dependencies import DependencySchema, get_dependencies

# pydantic_yaml won't be in the layer
try:
    from pydantic_yaml import YamlModel  # type: ignore
except ImportError:  # pragma: no cover

    class YamlModel(BaseModel):  # type: ignore
        ...


LHReturnType = TypeVar("LHReturnType")
LH = Callable[..., LHReturnType]  # Lambda Handler
Decorator = Callable[[LH], LH]


class Methods(str, Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PUT = "put"
    PATCH = "patch"


DEFAULT_STATUS_CODES = {
    Methods.GET: "200",
    Methods.POST: "201",
    Methods.DELETE: "204",
    Methods.PUT: "200",
    Methods.PATCH: "200",
}


class Functions(str, Enum):
    LAYER = "layer"
    OPENAPI = "openapi"
    SLS = "sls"


class SpyRoute(BaseModel):
    name: str
    path: str
    method: Methods
    handler: LH
    status_code: int
    tags: list[str] | None = Field(None)
    summary: str = Field("API endpoint")
    description: str | None = Field(None)
    request_body_arg_name: str | None = Field(None)
    request_body: type[BaseModel] | None = Field(None)
    response_class: type[BaseModel] | None = Field(None)
    header_params: list[ParamSchema] = Field(default_factory=list)
    path_params: list[ParamSchema] = Field(default_factory=list)
    query_params: list[ParamSchema] = Field(default_factory=list)
    # dependencies: list[DependencySchema] = Field(default_factory=list)
    use_vpc: bool = Field(True)
    authorizer: str | None = Field(None)
    layers: list[str] = Field(default_factory=list)
    add_event: bool = Field(default=False)
    add_context: bool = Field(default=False)
    skip_validation: bool = Field(default=False)

    @model_validator(mode="before")
    def set_status_code(  # type: ignore
        cls: type[Self],  # noqa: N805
        values: dict[str, Any],
    ) -> dict[str, Any]:
        if values.get("status_code") is None:
            values["status_code"] = DEFAULT_STATUS_CODES[values["method"]]

        return values

    @field_validator("summary", mode="before")
    def set_summary(cls: type[Self], summary: str | None) -> str:  # type: ignore  # noqa: N805
        return summary or "API endpoint"

    @model_validator(mode="after")
    def validate_handler_params(  # type: ignore
        cls: type[Self],  # noqa: N805
        model: Self,
    ) -> dict[str, Any]:
        method: Methods = model.method
        path: str = model.path
        handler: LH = model.handler

        path_params = get_path_param_names(path)
        handler_args = resolve_handler_args(handler)
        # values["dependencies"] = get_dependencies(handler)

        args = inspect.signature(handler).parameters
        args_count = len(args)
        # exclude lambdas event and context from count
        if "event" in args:
            model.add_event = True
            args_count -= 1
        if "context" in args:
            model.add_context = True
            args_count -= 1
        if handler_args.count != args_count and not model.skip_validation:
            msg = f'Unrecognized params for {method.upper()} method on "{path}" path!'
            raise RouteDefinitionError(msg)

        for path_param in path_params:
            if path_param not in handler_args.path.keys():
                msg = f"You did not specify {path_param} in your handler arguments!"
                raise RouteDefinitionError(msg)

        for path_arg in handler_args.path.keys():
            if path_arg not in path_params:
                msg = f'Your {path_arg} path parameter is missing in {method.upper()} method on "{path}" path!'
                raise RouteDefinitionError(msg)

        if method in (Methods.GET, Methods.DELETE) and handler_args.request_body:
            msg = f'{method.upper()} method on "{path}" cannot have request body!'
            raise RouteDefinitionError(msg)

        if handler_args.request_body:
            model.request_body = handler_args.request_body
            model.request_body_arg_name = handler_args.request_body_arg_name

        for attr_name in ("path", "query", "header"):
            setattr(
                model,
                f"{attr_name}_params",
                getattr(model, f"{attr_name}_params")
                + [param for _, param in getattr(handler_args, attr_name).items()],
            )

        return model


class _CloudFormationRef(BaseModel):
    stack_name: str
    export_name: str

    def __str__(self: Self) -> str:
        return f"${{cf:{self.stack_name}-${{opt:stage}}.{self.export_name}}}"


class CloudFormationRef(_CloudFormationRef):
    def __new__(cls: type[Self], stack_name: str, export_name: str) -> str:  # type: ignore
        instance = _CloudFormationRef(stack_name=stack_name, export_name=export_name)
        return str(instance)


class _JSONFileRef(BaseModel):
    file_path: str
    field: str

    def __str__(self: Self) -> str:
        return f"${{file({self.file_path}):{self.field}}}"


class JSONFileRef(_JSONFileRef):
    def __new__(cls: type[Self], file_path: str, field: str) -> str:  # type: ignore
        instance = _JSONFileRef(file_path=file_path, field=field)
        return str(instance)


def build_cognito_issue_url(user_pool_id: str | CloudFormationRef | JSONFileRef) -> str:
    return f"https://cognito-idp.${{region}}.amazonaws.com/{user_pool_id}"


class Authorizer(BaseModel):
    type: Literal["jwt"] = "jwt"  # noqa: A003
    identitySource: str = "$request.header.Authorization"  # noqa: N815
    issuerUrl: str  # noqa: N815
    audience: list[str | CloudFormationRef | JSONFileRef]


class CORS(BaseModel):
    allowedHeaders: list[str]  # noqa: N815
    exposedResponseHeaders: list[str]  # noqa: N815
    allowedMethods: list[str]  # noqa: N815
    allowedOrigins: list[str]  # noqa: N815


class HTTPApi(BaseModel):
    authorizers: dict[str, Authorizer]
    cors: CORS


class VPC(BaseModel):
    securityGroupIds: list[str | CloudFormationRef | JSONFileRef]  # noqa: N815
    subnetIds: list[str | CloudFormationRef | JSONFileRef]  # noqa: N815


class Function(BaseModel):
    handler: str
    module: str
    events: list[dict[str, Any]]
    layers: list[str | CloudFormationRef | JSONFileRef]
    environment: dict[str, Any] | None = Field(None)

    @staticmethod
    def generate_rel_path_for_function(route: SpyRoute) -> str:
        return os.path.relpath(Path(route.handler.__code__.co_filename), Path().resolve())

    @classmethod
    def from_route(cls: type[Self], *, route: SpyRoute, path: str, method: Methods) -> Self:  # type: ignore
        rel_path = cls.generate_rel_path_for_function(route)
        http_api_event: dict[str, Any] = {"path": path, "method": method.upper()}
        if route.authorizer:
            http_api_event["authorizer"] = {"name": route.authorizer}

        return cls(
            handler=(f'{rel_path.split(os.sep)[-1].replace(".py", "")}' f".{route.handler.__name__}"),  # noqa: ISC001
            module="/".join(rel_path.split(os.sep)[:-1]),
            events=[{"httpApi": http_api_event}],
            layers=list(
                set(
                    [
                        CloudFormationRef(
                            stack_name="spy-layer",
                            export_name="ServerlesspyLayerExport",
                        )
                    ],
                    *route.layers,  # type: ignore
                )
            ),
        )


class Provider(BaseModel):
    name: str = "aws"
    runtime: str = "python3.9"
    region: str
    role: str | CloudFormationRef | JSONFileRef
    httpApi: HTTPApi  # noqa: N815
    vpc: VPC | None = Field(None)


class ServerlessConfig(YamlModel):
    service: str
    custom: dict[str, Any] | None = Field(None)
    plugins: list[str]
    configValidationMode: str = "error"  # noqa: N815
    provider: Provider
    package: dict[str, bool] = Field({"individually": True})
    functions: dict[str, Function] | None = Field(None)

    @field_validator("plugins", mode="before")
    def set_default_plugins(cls: type[Self], value: list[str] | None) -> list[str]:  # type: ignore  # noqa: N805
        value = value if value is not None else []
        value += [
            "serverless-python-requirements",
            "serverless-plugin-common-excludes",
            "serverless-plugin-include-dependencies",
        ]
        return list(set(value))



import datetime
import json
from decimal import Decimal
from typing import Any
from uuid import UUID


class JSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, datetime.datetime | datetime.date):
            return o.isoformat()
        elif isinstance(o, UUID):
            return str(o)
        elif isinstance(o, Decimal):
            return float(o)
        return super().default(o)  # pragma: no cover



"""ServerlessPy package ;D"""

__version__ = "0.2.0"

from aws_spy import responses
from aws_spy.core.logging import logger
from aws_spy.core.params_alias import Header, Path, Query
from aws_spy.core.schemas import (
    CORS,
    VPC,
    Authorizer,
    CloudFormationRef,
    HTTPApi,
    JSONFileRef,
    Provider,
    ServerlessConfig,
    build_cognito_issue_url,
)
from aws_spy.main import SpyAPI, SpyRouter

__all__ = (
    "SpyAPI",
    "SpyRouter",
    "Query",
    "Path",
    "Header",
    "ServerlessConfig",
    "Provider",
    "VPC",
    "HTTPApi",
    "CORS",
    "Authorizer",
    "build_cognito_issue_url",
    "JSONFileRef",
    "CloudFormationRef",
    "logger",
    "responses",
)



import pytest
from pydantic import BaseModel

from aws_spy import SpyAPI
from aws_spy.core.schemas import Methods


class Request(BaseModel):
    ...


@pytest.mark.parametrize("method", [method for method in Methods if method])
def test_request_body(app: SpyAPI, method: Methods) -> None:
    app_method = getattr(app, method.value)
    path = "/somepath"

    @app_method(path, "lambda", response_class=Request)
    def handler() -> None:
        ...

    assert len(app.routes) == 1
    route = app.routes[path][method]
    assert route.response_class == Request
